import numpy as np
import struct
from pathlib import Path
from datetime import datetime, timedelta

MAX_POSIX_TIME = datetime(2038, 12, 31)


class Detector:
    def __init__(self, filename):
        self.filename = Path(filename)
        self._data = None

        if self.filename.suffix == ".detx":
            with open(self.filename, "r") as fin:
                self._data = read_detx(fin)
        elif self.filename.suffix == ".datx":
            with open(self.filename, "rb") as fin:
                self._data = read_datx(fin)
        else:
            raise Exception(f"Unsupported file extension: {self.filename}")

    @property
    def detid(self):
        return self._data["detid"]

    @property
    def modules(self):
        return self._data["modules"]

    @property
    def pmts(self):
        return self._data["pmts"]

    @property
    def version(self):
        return self._data["version"]

    @property
    def utm(self):
        return (self._data["utm_position"], self._data["utm_ref_grid"])

    @property
    def n_modules(self):
        return self._data["n_modules"]

    @property
    def validity(self):
        return self._data["validity"]


def unix2datetime(timestamp):
    if timestamp > MAX_POSIX_TIME.timestamp():
        timestamp = MAX_POSIX_TIME.timestamp()
    return datetime(1970, 1, 1) + timedelta(seconds=timestamp)


def _readstring(io):
    length = np.frombuffer(io.read(4), dtype=np.int32)[0]
    return io.read(length).decode("utf-8")


def read_detx(io):
    lines = io.readlines()

    header_last = max(
        [i for i, line in enumerate(lines) if line.strip() and line.startswith("#")]
    )

    lines = [
        line.strip()
        for line in lines[header_last:]
        if line.strip() and not line.startswith("#")
    ]

    first_line = lines[0].lower()

    if "v" in first_line:
        det_id, version = map(int, first_line.split("v"))
        validity = tuple([unix2datetime(float(t)) for t in lines[1].split()])
        utm_parts = lines[2].split()
        utm_position = np.array(utm_parts[3:6], dtype=np.float64)
        utm_ref_grid = " ".join(utm_parts[1:3])
        n_modules = int(lines[3])
        idx = 4
    else:
        det_id, n_modules = map(int, first_line.split())
        version = 1
        utm_position = None
        utm_ref_grid = None
        validity = None
        idx = 1

    module_dtype = np.dtype(
        [
            ("module_id", np.int32),
            ("du", np.int32),
            ("floor", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
            ("qa", np.float64),
            ("qb", np.float64),
            ("qc", np.float64),
            ("qd", np.float64),
            ("t0", np.float64),
            ("status", np.int32),
            ("n_pmts", np.int32),
        ]
    )

    pmts_dtype = np.dtype(
        [
            ("module_id", np.int32),
            ("pmt_id", np.int32),
            ("channel_id", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
            ("dir_x", np.float64),
            ("dir_y", np.float64),
            ("dir_z", np.float64),
            ("t0", np.float64),
            ("status", np.int32),
        ]
    )

    modules_data = []
    pmts_data = []

    floor_counter = 1
    last_string = -1

    for _ in range(n_modules):
        elements = lines[idx].split()
        module_id, string, floor = map(int, elements[:3])

        if floor == -1:
            if last_string == -1:
                last_string = string
            elif last_string != string:
                floor_counter = 1
                last_string = string
            floor = floor_counter
            floor_counter += 1

        if version >= 4:
            module_pos = np.array(elements[3:6], dtype=np.float64)
            q = np.array(elements[6:10], dtype=np.float64)
            t0 = float(elements[10])
        else:
            module_pos = np.zeros(3, dtype=np.float64)
            q = np.zeros(4, dtype=np.float64)
            t0 = 0.0

        status = float(elements[11]) if version >= 5 else 0.0
        n_pmts = int(elements[-1])

        pmts = np.zeros(n_pmts, dtype=pmts_dtype)

        for i in range(n_pmts):
            pmt_data = lines[idx + i + 1].split()
            pmts_data.append(
                (
                    int(module_id),
                    int(pmt_data[0]),
                    i,
                    *np.array(pmt_data[1:4], dtype=np.float64),
                    *np.array(pmt_data[4:7], dtype=np.float64),
                    float(pmt_data[7]),
                    int(pmt_data[8]) if version >= 3 else 0,
                )
            )

        modules_data.append(
            (module_id, string, floor, *module_pos, *q, t0, status, n_pmts)
        )

        idx += n_pmts + 1

    return {
        "version": version,
        "detid": det_id,
        "validity": validity,
        "utm_position": utm_position,
        "utm_ref_grid": utm_ref_grid,
        "n_modules": n_modules,
        "modules": np.array(modules_data, dtype=module_dtype),
        "pmts": np.array(pmts_data, dtype=pmts_dtype),
    }


def read_datx(io):
    comment_marker = b"####"
    supported_versions = {5}

    comments = []
    while io.read(4) == comment_marker:
        comments.append(_readstring(io))
    io.seek(max(0, io.tell() - len(comment_marker)))

    det_id = struct.unpack("i", io.read(4))[0]
    version = int(_readstring(io)[1:])
    if version not in supported_versions:
        raise ValueError(
            f"DATX version {version} is not supported. Supported versions: {supported_versions}"
        )

    validity = (
        unix2datetime(struct.unpack("d", io.read(8))[0]),
        unix2datetime(struct.unpack("d", io.read(8))[0]),
    )
    _readstring(io)  # Ignoring "UTM"
    wgs = _readstring(io)
    zone = _readstring(io)
    utm_ref_grid = f"{wgs} {zone}"
    utm_position = np.array(struct.unpack("ddd", io.read(24)), dtype=np.float64)
    n_modules = struct.unpack("i", io.read(4))[0]

    module_dtype = np.dtype(
        [
            ("module_id", np.int32),
            ("du", np.int32),
            ("floor", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
            ("qa", np.float64),
            ("qb", np.float64),
            ("qc", np.float64),
            ("qd", np.float64),
            ("t0", np.float64),
            ("status", np.int32),
            ("n_pmts", np.int32),
        ]
    )

    pmts_dtype = np.dtype(
        [
            ("module_id", np.int32),
            ("pmt_id", np.int32),
            ("channel_id", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
            ("dir_x", np.float64),
            ("dir_y", np.float64),
            ("dir_z", np.float64),
            ("t0", np.float64),
            ("status", np.int32),
        ]
    )

    modules_data = []
    pmts_data = []

    for _ in range(n_modules):
        module_id = struct.unpack("i", io.read(4))[0]
        du, floor = struct.unpack("ii", io.read(8))
        module_pos = struct.unpack("ddd", io.read(24))
        q = struct.unpack("dddd", io.read(32))
        module_t0 = struct.unpack("d", io.read(8))[0]
        module_status = struct.unpack("i", io.read(4))[0]
        n_pmts = struct.unpack("i", io.read(4))[0]

        for channel in range(n_pmts):
            pmt_id = struct.unpack("i", io.read(4))[0]
            pmt_pos = struct.unpack("ddd", io.read(24))
            pmt_dir = struct.unpack("ddd", io.read(24))
            pmt_t0 = struct.unpack("d", io.read(8))[0]
            pmt_status = struct.unpack("i", io.read(4))[0]
            pmts_data.append(
                (module_id, pmt_id, channel, *pmt_pos, *pmt_dir, pmt_t0, pmt_status)
            )

        modules_data.append(
            (module_id, du, floor, *module_pos, *q, module_t0, module_status, n_pmts)
        )

    return {
        "version": version,
        "detid": det_id,
        "validity": validity,
        "utm_position": utm_position,
        "utm_ref_grid": utm_ref_grid,
        "n_modules": n_modules,
        "modules": np.array(modules_data, dtype=module_dtype),
        "pmts": np.array(pmts_data, dtype=pmts_dtype),
    }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
from km3net_testdata import data_path
from km3det import Detector


class TestDetectorV5(unittest.TestCase):
    def setUp(self):
        self.datx_path = data_path("datx/KM3NeT_00000133_20221025.datx")
        self.detx_path = data_path("detx/KM3NeT_00000133_20221025.detx")

    def test_datx_reader(self):
        self.datx = Detector(self.datx_path)
        assert self.datx.n_modules == 399
        assert self.datx.n_modules == len(self.datx.modules)
        assert len(self.datx.pmts) == 11718
        assert self.datx.version == 5

    def test_detx_reader(self):
        self.detx = Detector(self.detx_path)
        assert self.detx.n_modules == 399
        assert self.detx.n_modules == len(self.detx.modules)
        assert len(self.detx.pmts) == 11718
        assert self.detx.version == 5



"""
How to read PMT parameters file
===============================

The following example shows how to read detector file using the ``Detector`` class
"""

from km3det import PMT_parameters
from km3net_testdata import data_path

pmt_file_path = data_path("pmt/calibration_00000117_H_1.0.0_00013757_00013826_1.txt")

#####################################################
# Initialising a Detector
# -----------------------
# Detector take a file path as input:


pars = PMT_parameters(pmt_file_path)


#####################################################
# Look at PMTs level information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The PMTs level information can be accessed this way
# (just printing the first 10 rows here):

print(pars.pmts[:10])

#####################################################
# The name of each fields can be obtained like so:

print(pars.pmts.dtype.names)



import numpy as np
from pathlib import Path


class PMT_parameters:
    def __init__(self, filename):
        self.filename = Path(filename)
        self._data = None

        if self.filename.suffix == ".txt":
            with open(self.filename, "r") as fin:
                self._data = read_txt_pmt_parameters(fin)

    @property
    def pmts(self):
        return self._data["pmts"]


def read_txt_pmt_parameters(io):
    lines = io.readlines()

    lines = [line.strip() for line in lines if line.strip() and line.startswith("PMT=")]

    pmts_dtype = np.dtype(
        [
            ("module_id", np.int32),
            ("channel_id", np.int32),
            ("QE", np.float64),
            ("gain", np.float64),
            ("gain_spread", np.float64),
            ("rise_time", np.float64),
            ("tts", np.float64),
            ("threshold", np.float64),
        ]
    )
    pmts_data = []

    for line in lines:
        elements = line.split()
        pmts_data.append(tuple(elements[1:]))

    return {"pmts": np.array(pmts_data, dtype=pmts_dtype)}



"""
How to read detector file
=========================

The following example shows how to read detector file using the ``Detector`` class
"""

from km3det import Detector
from km3net_testdata import data_path

datx_file_path = data_path("datx/KM3NeT_00000133_20221025.datx")

#####################################################
# Initialising a Detector
# -----------------------
# Detector take a file path as input:


det = Detector(datx_file_path)

#####################################################
# Look at module level information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The (optical) modules level information can be accessed this way
# (just printing the first 10 rows here):

print(det.modules[:10])

#####################################################
# The name of each fields can be obtained like so:

print(det.modules.dtype.names)


#####################################################
# Look at PMTs level information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The PMTs level information can be accessed this way
# (just printing the first 10 rows here):

print(det.pmts[:10])

#####################################################
# The name of each fields can be obtained like so:

print(det.pmts.dtype.names)



try:
    from importlib.metadata import version as get_version

    version = get_version(__name__)
except ImportError:
    from pkg_resources import get_distribution

    version = get_distribution(__name__).version

from .detector import Detector
from .pmt_parameters import PMT_parameters



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
from km3net_testdata import data_path
from km3det import PMT_parameters


class TestPMTparameters(unittest.TestCase):
    def setUp(self):
        self.pars_path = data_path(
            "pmt/calibration_00000117_H_1.0.0_00013757_00013826_1.txt"
        )

    def test_reader(self):
        self.pars = PMT_parameters(self.pars_path)
        assert len(self.pars.pmts) == 7254
